@@ -1,47 +1,44 @@
+# -*- coding: utf-8 -*-
+
+from __future__ import (absolute_import, division, print_function,
+                        unicode_literals)
 import re
-from collections import defaultdict
-from cStringIO import StringIO
+from codecs import open
 from xml import sax
-from htmllib import HTMLParser
 from xml.dom import minidom as dom
+from nine import IS_PYTHON2, basestring, str, iteritems
+
+if IS_PYTHON2:
+    from cStringIO import StringIO as BytesIO
+else:
+    from io import BytesIO
 
 from . import ir
 from . import template
+from .ddict import defaultdict
+from .doctype import DocumentTypeDeclaration, extract_dtd
+from .entities import html5, unescape
+from .html_utils import (HTML_OPTIONAL_END_TAGS, HTML_REQUIRED_END_TAGS,
+                         HTML_CDATA_TAGS)
 from .markup_template import QDIRECTIVES, QDIRECTIVES_DICT
-from .html_utils import HTML_OPTIONAL_END_TAGS
 
 impl = dom.getDOMImplementation(' ')
 
-_pattern = r'''
-\$(?:
-    (?P<expr_escaped>\$) |      # Escape $$
-    (?P<expr_named>[_a-z][_a-z0-9.]*) | # $foo.bar
-    {(?P<expr_braced>) | # ${....
-    (?P<expr_invalid>)
-)'''
-_re_pattern = re.compile(_pattern, re.VERBOSE | re.IGNORECASE|re.MULTILINE)
 
-def XMLTemplate(
-    source=None,
-    filename=None,
-    **kw):
-    if 'mode' in kw:
-        mode = kw['mode']
-        force_mode = True
-    else:
-        mode = 'xml'
-        force_mode = False
-    is_fragment = kw.pop('is_fragment', False)
+def XMLTemplate(source=None, filename=None, mode=None, is_fragment=False,
+                encoding='utf-8'):
     if source is None:
-        source = open(filename).read()
+        with open(filename, encoding=encoding) as f:
+            source = f.read()  # source is a unicode string
     if filename is None:
         filename = '<string>'
     doc = _Parser(filename, source).parse()
     expand(doc)
-    compiler = _Compiler(filename, doc, mode, is_fragment, force_mode)
+    compiler = _Compiler(filename, doc, mode=mode, is_fragment=is_fragment)
     ir_ = compiler.compile()
     return template.from_ir(ir_)
 
+
 def annotate(gen):
     def inner(self, node, *args, **kwargs):
         for x in gen(self, node, *args, **kwargs):
@@ -49,52 +46,46 @@ def annotate(gen):
             yield x
     return inner
 
-class _Compiler(object):
-    mode_lookup = {
-        'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd':'xml',
-        'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd':'xml',
-        'http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd':'xml',
-        'http://www.w3.org/TR/html4/strict.dtd':'html',
-        'http://www.w3.org/TR/html4/loose.dtd':'html',
-        'http://www.w3.org/TR/html4/frameset.dtd':'html',
-        }
 
-    def __init__(self, filename, doc, mode='xml', is_fragment=False, force_mode=False):
+class _Compiler(object):
+    def __init__(self, filename, doc, mode=None, is_fragment=False):
         self.filename = filename
         self.doc = doc
-        self.mode = mode
+        self.is_fragment = is_fragment
         self.functions = defaultdict(list)
         self.functions['__main__()'] = []
         self.function_lnos = {}
         self.mod_py = []
         self.in_def = False
         self.is_child = False
-        self.is_fragment = is_fragment
-        if not force_mode and self.doc.doctype:
-            if self.doc.doctype.toxml().lower() == '<!doctype html>':
-                self.mode = 'html5'
-            elif self.doc.doctype.systemId is None:
-                self.mode = 'html'
-            else:
-                self.mode = self.mode_lookup.get(
-                    self.doc.doctype.systemId, 'xml')
+        # The rendering mode is either specified in the *mode* argument,
+        # or inferred from the DTD:
+        self._dtd = DocumentTypeDeclaration.matching(self.doc._dtd)
+        if mode:
+            self.mode = mode
+        elif self._dtd:
+            self.mode = self._dtd.rendering_mode
+        else:  # The template might contain an unknown DTD
+            self.mode = 'xml'  # by default
 
     def compile(self):
         body = list(self._compile_node(self.doc.firstChild))
-        if not self.is_fragment and not self.is_child: # Never emit doctypes on fragments
-            if self.mode == 'xml' and self.doc.doctype:
-                dt = ir.TextNode(self.doc.doctype.toxml())
-                dt.filename = self.filename
-                dt.lineno = 1
-                body.insert(0, dt)
+        # Never emit doctypes on fragments
+        if not self.is_fragment and not self.is_child:
+            if self.doc._dtd:
+                dtd = self.doc._dtd
             elif self.mode == 'html5':
-                dt = ir.TextNode('<!DOCTYPE html>')
-                dt.filename = self.filename
-                dt.lineno = 1
-                body.insert(0, dt)
+                dtd = '<!DOCTYPE html>'
+            else:
+                dtd = None
+            if dtd:
+                dtd = ir.TextNode(dtd)
+                dtd.filename = self.filename
+                dtd.lineno = 1
+                body.insert(0, dtd)
         self.functions['__main__()'] = body
         defs = []
-        for k,v in self.functions.iteritems():
+        for k, v in iteritems(self.functions):
             node = ir.DefNode(k, *v)
             node.lineno = self.function_lnos.get(k)
             defs.append(node)
@@ -104,7 +95,8 @@ class _Compiler(object):
         return node
 
     def _anno(self, dom_node, ir_node):
-        if ir_node.lineno: return
+        if ir_node.lineno:
+            return
         ir_node.filename = self.filename
         ir_node.lineno = dom_node.lineno
 
@@ -117,8 +109,9 @@ class _Compiler(object):
             return self._compile_pi(node)
         elif node.tagName.startswith('py:'):
             # Handle directives
-            compiler = getattr(self, '_compile_%s' % node.tagName.split(':')[-1],
-                               self._compile_xml)
+            compiler = getattr(
+                self, '_compile_%s' % node.tagName.split(':')[-1],
+                self._compile_xml)
             return compiler(node)
         else:
             return self._compile_xml(node)
@@ -127,14 +120,17 @@ class _Compiler(object):
     def _compile_xml(self, node):
         content = attrs = guard = None
         if node.hasAttribute('py:strip'):
-            guard = 'not (%s)' % node.getAttribute('py:strip')
+            guard = node.getAttribute('py:strip')
+            if guard == '':  # py:strip="" means yes, do strip the tag
+                guard = 'False'
+            else:
+                guard = 'not (%s)' % guard
             node.removeAttribute('py:strip')
-        yield ir.TextNode(u'<%s' % node.tagName, guard)
-        for k,v in node.attributes.items():
+        yield ir.TextNode('<%s' % node.tagName, guard)
+        for k, v in sorted(node.attributes.items()):
             tc = _TextCompiler(self.filename, v, node.lineno,
-                               ir.TextNode)
+                               ir.TextNode, in_html_attr=True)
             v = list(tc)
-            # v = u''.join(n.text for n in tc)
             if k == 'py:content':
                 content = node.getAttribute('py:content')
                 continue
@@ -145,26 +141,41 @@ class _Compiler(object):
         if attrs:
             yield ir.AttrsNode(attrs, guard, self.mode)
         if content:
-            yield ir.TextNode(u'>', guard)
+            yield ir.TextNode('>', guard)
             yield ir.ExprNode(content)
-            yield ir.TextNode(u'</%s>' % node.tagName, guard)
+            yield ir.TextNode('</%s>' % node.tagName, guard)
         else:
             if node.childNodes:
-                yield ir.TextNode(u'>', guard)
-                for cn in node.childNodes:
-                    for x in self._compile_node(cn):
-                        yield x
+                yield ir.TextNode('>', guard)
+                if node.tagName in HTML_CDATA_TAGS:
+                    # Special behaviour for <script>, <style> tags:
+                    if self.mode == 'xml':  # Start escaping
+                        yield ir.TextNode('/*<![CDATA[*/')
+                    # Need to unescape the contents of these tags
+                    for child in node.childNodes:
+                        assert isinstance(child, dom.Text)
+                        for x in self._compile_text(child):
+                            x.text = unescape(x.text)
+                            yield x
+                    if self.mode == 'xml':  # Finish escaping
+                        yield ir.TextNode('/*]]>*/')
+                else:
+                    for cn in node.childNodes:
+                        for x in self._compile_node(cn):
+                            yield x
                 if not (self.mode.startswith('html')
                         and node.tagName in HTML_OPTIONAL_END_TAGS):
-                    yield ir.TextNode(u'</%s>' % node.tagName, guard)
+                    yield ir.TextNode('</%s>' % node.tagName, guard)
+            elif node.tagName in HTML_REQUIRED_END_TAGS:
+                yield ir.TextNode('></%s>' % node.tagName, guard)
             else:
                 if self.mode.startswith('html'):
-                    if  node.tagName in HTML_OPTIONAL_END_TAGS:
-                        yield ir.TextNode(u'>', guard)
+                    if node.tagName in HTML_OPTIONAL_END_TAGS:
+                        yield ir.TextNode('>', guard)
                     else:
-                        yield ir.TextNode(u'></%s>' % node.tagName, guard)
+                        yield ir.TextNode('></%s>' % node.tagName, guard)
                 else:
-                    yield ir.TextNode(u'/>', guard)
+                    yield ir.TextNode('/>', guard)
 
     @annotate
     def _compile_replace(self, node):
@@ -206,11 +217,12 @@ class _Compiler(object):
         decl = fname + '()'
         body = list(self._compile_nop(node))
         if not body:
-            body = [ ir.PassNode() ]
+            body = [ir.PassNode()]
         self.functions[decl] = body
         if self.is_child:
             parent_block = 'parent.' + fname
-            body.insert(0, ir.PythonNode(ir.TextNode('parent_block=%s' % parent_block)))
+            body.insert(0,
+                ir.PythonNode(ir.TextNode('parent_block=%s' % parent_block)))
         else:
             yield ir.ExprNode(decl)
 
@@ -259,14 +271,14 @@ class _Compiler(object):
     @annotate
     def _compile_switch(self, node):
         # Filter out text nodes
-        body = [ x for x in self._compile_nop(node)
-                 if not isinstance(x, ir.TextNode) ]
+        body = [x for x in self._compile_nop(node)
+                if not isinstance(x, ir.TextNode)]
         yield ir.SwitchNode(node.getAttribute('test'), *body)
 
     @annotate
     def _compile_case(self, node):
         yield ir.CaseNode(node.getAttribute('value'),
-                         *list(self._compile_nop(node)))
+                          *list(self._compile_nop(node)))
 
     @annotate
     def _compile_if(self, node):
@@ -284,17 +296,22 @@ class _Compiler(object):
             for x in self._compile_node(c):
                 yield x
 
-class _TextCompiler(object):
 
+class _TextCompiler(object):
+    '''Separates expressions such as ${some_var} from the ordinary text
+    around them in the template source and generates ExprNode instances and
+    TextNode instances accordingly.
+    '''
     def __init__(self, filename, source, lineno,
-                 node_type=ir.TranslatableTextNode):
+                 node_type=ir.TranslatableTextNode, in_html_attr=False):
         self.filename = filename
         self.source = source
         self.orig_lineno = lineno
         self.lineno = 0
         self.pos = 0
         self.node_type = node_type
-        
+        self.in_html_attr = in_html_attr
+
     def text(self, text):
         node = self.node_type(text)
         node.lineno = self.real_lineno
@@ -302,7 +319,9 @@ class _TextCompiler(object):
         return node
 
     def expr(self, text):
-        node = ir.ExprNode(text)
+        # *safe* being True here avoids escaping twice, since
+        # HTML attributes are always escaped in the end.
+        node = ir.ExprNode(text, safe=self.in_html_attr)
         node.lineno = self.real_lineno
         self.lineno += text.count('\n')
         return node
@@ -311,9 +330,19 @@ class _TextCompiler(object):
     def real_lineno(self):
         return self.orig_lineno + self.lineno
 
+    _pattern = r'''
+    \$(?:
+        (?P<expr_escaped>\$) |      # Escape $$
+        (?P<expr_named>[_a-z][_a-z0-9.]*) | # $foo.bar
+        {(?P<expr_braced>) | # ${....
+        (?P<expr_invalid>)
+    )'''
+    _re_pattern = re.compile(
+        _pattern, re.VERBOSE | re.IGNORECASE | re.MULTILINE)
+
     def __iter__(self):
         source = self.source
-        for mo in _re_pattern.finditer(source):
+        for mo in self._re_pattern.finditer(source):
             start = mo.start()
             if start > self.pos:
                 yield self.text(source[self.pos:start])
@@ -331,8 +360,8 @@ class _TextCompiler(object):
             else:
                 msg = 'Syntax error %s:%s' % (self.filename, self.real_lineno)
                 for i, line in enumerate(self.source.splitlines()):
-                    print '%3d %s' % (i+1, line)
-                print msg
+                    print('%3d %s' % (i + 1, line))
+                print(msg)
                 assert False, groups
         if self.pos != len(source):
             yield self.text(source[self.pos:])
@@ -340,19 +369,38 @@ class _TextCompiler(object):
     def _get_braced_expr(self):
         try:
             compile(self.source[self.pos:], '', 'eval')
-        except SyntaxError, se:
-            end = se.offset+self.pos
-            text = self.source[self.pos:end-1]
+        except SyntaxError as se:
+            end = se.offset + self.pos
+            text = self.source[self.pos:end - 1]
             self.pos = end
             return self.expr(text)
-    
+
+
 class _Parser(sax.ContentHandler):
+    DTD = '<!DOCTYPE kajiki SYSTEM "kajiki.dtd">'
 
     def __init__(self, filename, source):
-        self._filename = filename
-        self._source = source
-        self._doc = None
+        '''XML defines only a few entities; HTML defines many more.
+        The XML parser errors out when it finds HTML entities, unless the
+        template contains a reference to an external DTD (in which case
+        skippedEntity() gets called, this is what we want). In other words,
+        we want to trick expat into parsing XML + HTML entities for us.
+        We wouldn't force our users to declare everyday HTML entities!
+
+        So, for the parsing stage, we detect the doctype in the template and
+        replace it with our own; then in the compiling stage we put the
+        user's doctype back in. The XML parser is thus tricked and nobody
+        needs to know this implementation detail of Kajiki.
+        '''
+        if not isinstance(source, str):
+            raise TypeError('The template source must be a unicode string.')
         self._els = []
+        self._doc = dom.Document()
+        self._filename = filename
+        # Store the original DTD in the document for the compiler to use later
+        self._doc._dtd, position, source = extract_dtd(source)
+        # Use our own DTD just for XML parsing
+        self._source = source[:position] + self.DTD + source[position:]
 
     def parse(self):
         self._parser = parser = sax.make_parser()
@@ -361,19 +409,28 @@ class _Parser(sax.ContentHandler):
         parser.setFeature(sax.handler.feature_namespaces, False)
         parser.setProperty(sax.handler.property_lexical_handler, self)
         parser.setContentHandler(self)
-        parser.parse(StringIO(self._source))
+        source = sax.xmlreader.InputSource()
+        # Sweet XMLReader.parse() documentation says:
+        # "As a limitation, the current implementation only accepts byte
+        # streams; processing of character streams is for further study."
+        # So if source is unicode, we pre-encode it:
+        # TODO Is this dance really necessary? Can't I just call a function?
+        byts = self._source.encode('utf-8')
+        source.setEncoding('utf-8')
+        source.setByteStream(BytesIO(byts))
+        source.setSystemId(self._filename)
+        parser.parse(source)
         return self._doc
 
-    ## ContentHandler implementation
+    # ContentHandler implementation
     def startDocument(self):
-        self._doc = dom.Document()
         self._els.append(self._doc)
 
     def startElement(self, name, attrs):
         el = self._doc.createElement(name)
         el.lineno = self._parser.getLineNumber()
-        for k,v in attrs.items():
-            el.setAttribute(k,v)
+        for k, v in attrs.items():
+            el.setAttribute(k, v)
         self._els[-1].appendChild(el)
         self._els.append(el)
 
@@ -382,6 +439,7 @@ class _Parser(sax.ContentHandler):
         assert name == popped.tagName
 
     def characters(self, content):
+        content = sax.saxutils.escape(content)
         node = self._doc.createTextNode(content)
         node.lineno = self._parser.getLineNumber()
         self._els[-1].appendChild(node)
@@ -392,20 +450,26 @@ class _Parser(sax.ContentHandler):
         self._els[-1].appendChild(node)
 
     def skippedEntity(self, name):
-        content = unicode(HTMLParser.entitydefs[name], 'latin-1')
-        return self.characters(content)
+        '''Deals with an HTML entity such as &nbsp;
+        (XML itself defines very few entities.)
 
-    def startElementNS(self, name, qname, attrs): # pragma no cover
-        raise NotImplementedError, 'startElementNS'
+        The presence of a SYSTEM doctype makes expat say "hey, that MIGHT be
+        a valid entity, better pass it along to sax and find out!"
+        (Since expat is nonvalidating, it never reads the external doctypes.)
+        '''
+        return self.characters(html5[name])
 
-    def endElementNS(self, name, qname):# pragma no cover
-        raise NotImplementedError, 'startElementNS'
+    def startElementNS(self, name, qname, attrs):  # pragma no cover
+        raise NotImplementedError('startElementNS')
 
-    def startPrefixMapping(self, prefix, uri):# pragma no cover
-        raise NotImplemented, 'startPrefixMapping'
+    def endElementNS(self, name, qname):  # pragma no cover
+        raise NotImplementedError('startElementNS')
 
-    def endPrefixMapping(self, prefix):# pragma no cover
-        raise NotImplemented, 'endPrefixMapping'
+    def startPrefixMapping(self, prefix, uri):  # pragma no cover
+        raise NotImplementedError('startPrefixMapping')
+
+    def endPrefixMapping(self, prefix):  # pragma no cover
+        raise NotImplementedError('endPrefixMapping')
 
     # LexicalHandler implementation
     def comment(self, text):
@@ -413,11 +477,18 @@ class _Parser(sax.ContentHandler):
         node.lineno = self._parser.getLineNumber()
         self._els[-1].appendChild(node)
 
-    def startCDATA(self): pass
-    def endCDATA(self): pass
+    def startCDATA(self):
+        pass
+
+    def endCDATA(self):
+        pass
+
     def startDTD(self, name, pubid, sysid):
         self._doc.doctype = impl.createDocumentType(name, pubid, sysid)
-    def endDTD(self): pass
+
+    def endDTD(self):
+        pass
+
 
 def expand(tree, parent=None):
     if isinstance(tree, dom.Document):
@@ -438,7 +509,8 @@ def expand(tree, parent=None):
         tree.removeAttribute('py:extends')
         tree.childNodes.insert(0, el)
     for directive, attr in QDIRECTIVES:
-        if not tree.hasAttribute(directive): continue
+        if not tree.hasAttribute(directive):
+            continue
         value = tree.getAttribute(directive)
         tree.removeAttribute(directive)
         # nsmap = (parent is not None) and parent.nsmap or tree.nsmap
@@ -454,4 +526,3 @@ def expand(tree, parent=None):
     for child in tree.childNodes:
         expand(child, tree)
     return tree
-
